"""
Wrappers and functions used in different parts of the code.
"""
from datetime import datetime as dt


def timing(active: bool = True):
    """Decorator for timing with option to turn it off"""

    def decorator(func):
        def wrapper(inst, *args, **kwargs):
            if active:
                dt0 = dt.now()
                output = func(*args, **kwargs)
                dt1 = dt.now()
                tot_time = (dt1 - dt0).total_seconds()
                print(f"{func.__name__}: {tot_time:>{30 - len(func.__name__)}.3f}s")
            else:
                output = func(inst, *args, **kwargs)

            return output

        return wrapper

    return decorator


def check_attrs(*attrs):
    """
    Decorator that checks if the class instance has the attributes in `attrs`.
    If it doesn't, return the value as is rather than calculate it. So properties
    can "cache" their values if the largest arrays they rely on are deleted.
    """

    def decorator(func):
        def wrapper(inst, *args, **kwargs):
            for attr in attrs:
                if not hasattr(inst, attr):
                    # get hidden attribute name (has an underscore in the front)
                    return inst.__getattribute__(f"_{func.__name__}")
            return func(inst, *args, **kwargs)

        return wrapper

    return decorator



"""
File that holds the CVIntegrator class
"""

from copy import deepcopy
from itertools import product
from numbers import Number
from re import findall
from sys import getsizeof
from typing import Optional, Sequence, Union

import numpy as np
from nptyping import Float, NDArray, Shape
from numpy.random import RandomState
from vegas import Integrator

from ._types import _ftype
from ._wrappers import check_attrs, timing
from .functions import Function, make_func


def quick_integrate(
    function: _ftype,
    evals: int,
    tot_iters: int,
    bounds: Union[Sequence[tuple[float, float]], tuple[float, float]],
    cv_iters: Optional[Union[list[int], int, str]] = None,
    cv_means: Union[float, Sequence[float]] = 1,
    rng: Optional[RandomState] = None,
    cname: str = None,
    name: str = None,
    **params: float,
):
    """
    A convenience method if you don't want to create a custom Function class via
    make_func. This method will do it for you, pass the class to the CVIntegrator and
    run the integrate method, passing back the CVIntegrator.

    Parameters:
    function - The function to be integrated. Must be vectorized, see the docstring for
        make_func for more info (from control_vegas import make_func).
    evals - Number of Vegas evaluations per iteration (called `neval` by Vegas). This
        is the default value used by create_maps, get_is_cv_values but those can be
        specified separately.
    tot_iters - Total number of iterations for Vegas to do (called `nitn` by Vegas).
    bounds - The bounds of the integration for each dimension given as a list of tuples.
        The CVIntegrator class uses the dimension from the Function class so that the
        bounds argument is optional. But here it is opposite, the dimension of the
        function is implied from the number of bounds.
    cv_iters - List of iterations to use as CVs. See the docstring for CVIntegrator
        for more information about the options. Most simply can be an integer
        representing a single CV or a list of integers representing multiple.
    cv_means (default 1) - The value of E[g_i] but by the scheme laid out in
        `get_is_cv_values` to obtain the control variate, E[g_i] should be approximately
        one.
    rng (default None) - The Numpy Randomstate to use. If None, will create a new
        one.
    cname (default None) - The name of the class of the Function passed to CVIntegrator.
        If not specified, the __name__ attribute of `function` is passed capitalized.
    name (default None) - The name attribute of the Function. If not specified,
        the __name__ attribute is passed as is.
    params - Parameters for the function.
    """
    # make Function object
    function = make_func(
        cname=function.__name__.capitalize() if cname is None else cname,
        dimension=len(bounds),
        function=function,
        name=function.__name__ if name is None else name,
        **params,
    )
    # Create integrator object
    cvi = CVIntegrator(
        function=function,
        evals=evals,
        tot_iters=tot_iters,
        bounds=bounds,
        cv_iters=cv_iters,
        cv_means=cv_means,
        rng=rng,
    )
    # And integrate
    cvi.integrate()
    return cvi


# TODO: Automatically choose single CV: find the variance dip
class CVIntegrator:
    """
    Integrating a function f, we can equivalently integrate f'=f + c(g +E[g]) where the
    expectation value of g is known. Choosing the optimal value for c will necessarily
    reduce the variance. We can have an arbitrary number of g's (called control variates
    (CVs)). Here we use a previous Vegas adapted interation of f for g. So if we have a
    total of N iterations of f, then we can have a maximum of N-1 CVs.

    Below, f' is called self.weight_prime (or the 'CV function'), f is self.weight_value
    (or the 'adapted/IS function') and the CVs (the g_i's) are elements of the list
    self.cv_values (called CVs). These are the terminology used in the docstrings below.

    Explanation of parameters:
        neval - Passed on instantiation of class, check __init__ docstring for details.
        nitn - The number of iterations the Integrator is ran to get the adapted
            function.
        cv_nums - The number of control variates.
        map_neval - The number of evaluations per iteration when initially creating the
            map. Defaults to neval.
        jac_neval - The size of the Jacobian arrays, i.e. how finely split up the
            integration region is. Defaults to neval*nitn
        tot_neval - The total number of evaluations done when creating the fully adapted
            map.
    """

    # Should you print out the time it takes for the main 3 functions to run?
    TIMING = False
    # Memory threshold for when `memory="tiny"`
    TINY_THRESHOLD = 100

    def __init__(
        self,
        function: Function,
        evals: int,
        tot_iters: int,
        bounds: Optional[Sequence[tuple[float, float]]] = None,
        cv_iters: Optional[Union[list[int], int, str]] = None,
        cv_means: Union[float, Sequence[float]] = 1,
        rng: Optional[RandomState] = None,
        memory: str = "medium",
    ):
        """
        Takes in a Function class object from functions.py. One can make their own using
        the make_func function found in that file.

        Parameters:
        f - Function class with f to integrate.
        evals - Number of Vegas evaluations per iteration (called `neval` by Vegas).
            This is the default value used by create_maps, get_is_cv_values but those
            can be specified separately.
        tot_iters - Total number of iterations for Vegas to do (called `nitn` by Vegas).
        bounds (default None) - The bounds of the integration for each dimension. If not
            given, defaults to [0, 1] for every dimension.
        cv_iters (default None) - List of iterations to use as CVs. Defaults to no CVs.
            Can be passed as a single integer which is considered as a single control
            variate. Can also be passed as a string:
                - 'all': Use every iteration as a control variate.
                - 'all%n': Use every iteration mod n. For example, if tot_iters=10
                    and cv_iters='all%2', then it uses [2, 4, 6, 8]
                - 'all%n+b': Use every iteration (shifted by b) mod n. For example, if
                    tot_iters=10 and cv_iters='all%2+1', then use [1, 3, 5, 7, 9]
        cv_means (default 1) - The value of E[g_i] but by the scheme laid out in
            `get_is_cv_values` to obtain the control variate, E[g_i] should be
            approximately one.
        rng (default None) - The Numpy Randomstate to use. If None, will create a new
            one.
        memory (default 'medium') - Either 'low', 'medium', 'large' or `max. Determines
            what is saved. If `max`, save everything. If `large, don't save self.xs
            and self.is_jac. If 'medium', additionally don't save self.weight_value and
            self.weight_prime. If 'tiny', remove everything below the threshold
            TINY_TRESHOLD.
        """
        # Initialize private attributes for the properties
        self._init_results()

        self.function = function
        self.bounds = self.function.dim * [[0, 1]] if bounds is None else bounds
        self.neval = evals
        self.nitn = tot_iters

        self.cv_nitn = cv_iters
        # Create empty list if not specified, i.e. no control variates
        if self.cv_nitn is None:
            self.cv_nitn = []
        # If cv_iters is a number, put it into a list
        if isinstance(self.cv_nitn, int):
            self.cv_nitn = [self.cv_nitn]
        if isinstance(self.cv_nitn, str):
            # Find the mod and shift using regex
            all_str = findall(r"^all%(\d+)(?:\+(\d+))?$", self.cv_nitn)
            if all_str:
                # Extract those parameters (0 shift if not specified)
                mod = int(all_str[0][0])
                shift = 0 if not all_str[0][1] else int(all_str[0][1])
                # Create list according to those numbers
                shifted_cv_nitns = np.where(np.arange(self.nitn) % mod == 0)[0] + shift
                self.cv_nitn = list(shifted_cv_nitns[shifted_cv_nitns < self.nitn])
            elif self.cv_nitn == "all":
                self.cv_nitn = list(range(1, self.nitn))

        # Iteration 0 is no iteration at all, so remove it
        if 0 in self.cv_nitn:
            self.cv_nitn.remove(0)

        self.num_cvs = len(self.cv_nitn)
        self.cv_means = cv_means
        # A number implies a constant mean value
        if isinstance(self.cv_means, Number):
            self.cv_means = self.num_cvs * [cv_means]

        self.rng = RandomState() if rng is None else rng
        self.memory = memory

    def _init_results(self):
        """Initializes the private attributes for the listed properties."""
        for name, obj in self.__class__.__dict__.items():
            if isinstance(obj, property):
                self.__setattr__(f"_{name}", np.nan)

    @timing(active=TIMING)
    def create_maps(self, map_neval: Optional[int] = None) -> None:
        """
        Creates the maps corresponding to the adapted function, f, and the
        control variates, g_i.

        Parameters:
        map_neval (default None) - The number of evaluations per iteration as
            the maps are being created. Defaults to `self.neval`.
        """
        self.map_neval = self.neval if map_neval is None else map_neval
        integrator = Integrator(self.bounds)
        self._cv_maps = []
        self.tot_neval = 0

        # Do this if there actually are CVs, otherwise you don't need to
        if self.cv_nitn:
            result = integrator(
                self.function._f, nitn=self.cv_nitn[0], neval=self.map_neval
            )
            self._cv_maps.append(deepcopy(integrator.map))
            self.tot_neval += int(result.sum_neval)
            # For loop if there is more than 1 CV to save the others
            for cv_nitn_diff in np.diff(self.cv_nitn):
                result = integrator(
                    self.function._f, nitn=cv_nitn_diff, neval=self.map_neval
                )
                self._cv_maps.append(deepcopy(integrator.map))
                self.tot_neval += int(result.sum_neval)

            # And save the final map as the IS map
            result = integrator(
                self.function._f,
                nitn=self.nitn - self.cv_nitn[-1],
                neval=self.map_neval,
            )
            self._is_map = deepcopy(integrator.map)
            self.tot_neval += int(result.sum_neval)
        else:
            # Only have an IS map if there are no CVs
            result = integrator(self.function._f, nitn=self.nitn, neval=self.map_neval)
            self._is_map = deepcopy(integrator.map)
            self.tot_neval += int(result.sum_neval)

    @timing(active=TIMING)
    def get_is_cv_values(self, jac_neval: Optional[int] = None) -> None:
        """
        Calculates the  adapted function and the control variates from their maps.

        Parameters:
        jac_neval (default None) - The number of steps to split up `ys`, the unit
            hypercube, into. Defaults to `self.tot_neval`, the total number of
            iterations used when adapting the map.
        """
        self.jac_neval = self.neval * self.nitn if jac_neval is None else jac_neval

        # Uniformly distributed unit hypercube
        ys = self.rng.uniform(0, 1, (self.jac_neval, self.function.dim))
        # Find the Jacobian. If by importance sampling we transform f -> f/p, then
        # the Jacobian is 1/p
        xs = np.empty(ys.shape, float)
        is_jac = np.empty(ys.shape[0], float)
        self._is_map.map(ys, xs, is_jac)
        # The IS values
        self.weight_value = is_jac * self.function._f(xs)

        # Find the Jacobian(s) for the CV(s)
        self.cv_values, self.cv_jacs = [], []
        for cv_map in self._cv_maps:
            # Use inverse map for control variate to find CV Jacobian
            t_inv = np.empty(xs.shape, float)
            cv_jac = np.empty(xs.shape[0], float)
            cv_map.invmap(xs, t_inv, cv_jac)

            self.cv_values.append(is_jac / cv_jac)
            self.cv_jacs.append(cv_jac)

        if self.memory == "max":
            self.xs = xs
            self.is_jac = is_jac

    @timing(active=TIMING)
    def get_weight_prime(self, constant: bool = True) -> None:
        """
        Calculates the final CV function by finding the optimal coefficients
        for the control variates.

        Parameters:
        constant - If `True`, then the N optimal coefficients (for N CVs) will be
            constant for every `self.jac_neval` values in each function. This is much
            faster. If `False`, then calculate a different coefficient for each element.
            Explained in more detail in the `self._find_coefficients` docstring.
        """
        self._find_coefficients(constant=constant)
        self.weight_prime = self.weight_value + sum(
            [
                self.cs[ind] * (self.cv_values[ind] - self.cv_means[ind])
                for ind in range(self.num_cvs)
            ]
        )

    def _find_coefficients(self, constant: bool) -> None:
        """
        Finds the optimized values for the CV coefficients to minimize the variance via
        a matrix. Our equation to solve is of the form A=Bc where A and c are arrays and
        B is a matrix. We solve for c.

        Parameters:
        constant (default True) - Since the ith value of a control variate can be
            slightly correlated to its coefficient, we can remove said value to
            calculate the variance/covariance (and therefore the coefficient). Thus we
            can have different values (albeit similar ones) for each value. This is so
            if `constant=True`. If `constant=False`, only do single coefficient per CV.
        """
        # Create (num_cv, num_cv) matrix
        if constant:
            Bs = np.zeros((self.num_cvs, self.num_cvs))
        else:
            Bs = np.zeros((self.num_cvs, self.num_cvs, self.jac_neval))

        # Populate the B matrix
        for i, j in product(range(self.num_cvs), repeat=2):
            Bs[i, j] = self._cov(
                self.cv_values[i], self.cv_values[j], constant=constant
            )
        As = np.array(
            [
                -self._cov(self.weight_value, cv_value, constant=constant)
                for cv_value in self.cv_values
            ]
        )

        # Solve the system of equations for each index
        if constant:
            cs = np.linalg.solve(Bs, As)
        else:
            cs = np.array(
                [
                    np.linalg.solve(Bs[:, :, ind], As[:, ind])
                    for ind in range(self.jac_neval)
                ]
            )
        self.cs = cs.T

    def _cov(
        self,
        f1: NDArray[Shape["'*'"], Float],
        f2: NDArray[Shape["'*'"], Float],
        constant: bool,
    ) -> Union[float, NDArray[Shape["'*'"], Float]]:
        """
        Calculates the covariance between `f1` and `f2`.

        If `constant` is `False`: for each element of the array returned, the covariance
        is calculated with said element of `f1` and `f2` removed. This removes the
        potential correlation between the CV and its coefficient so that when
        calculating the expectation value of f', we can act linearly with
        E[c(g-E(g))]=0.

        If `constant` is `True`: return a single covariance.
        """
        if constant:
            return np.cov(f1, f2)[0, 1]

        prod = f1 * f2
        # For each index, calculate the covariance without the value
        # for said index to remove that bias otherwise E[x_prime] =/= E[x]
        cov = (np.sum(prod) - prod) / (self.jac_neval - 1) - (np.sum(f1) - f1) * (
            np.sum(f2) - f2
        ) / (self.jac_neval - 1) ** 2
        return cov

    def integrate(
        self,
        map_neval: Optional[int] = None,
        jac_neval: Optional[int] = None,
        constant: bool = True,
    ) -> None:
        """
        Runs the necessary functions to integrate the function in the order:
            1) self.create_maps
            2) self.get_is_cv_values
            3) self.get_weight_primes
        Check out the docstrings of these functions for more info on them.

        Parameters:
        map_neval (default None) - From self.create_maps docstring: The number of
            evaluations per iteration as the maps are being created. Defaults to
            `self.neval`.
        jac_neval (default None) - From self.get_is_cv_values docstring: The number of
            steps to split up `ys`, the unit hypercube, into. Defaults to
            `self.tot_neval`, the total number of iterations used when adapting the map.
        constant (default True) - From self.get_weight_primes: Since the ith value of a
            control variate can be slightly correlated to its coefficient, we can remove
            said value to calculate the variance/covariance (and therefore the
            coefficient). Thus we can have different values (albeit similar ones) for
            each value. This is so if `constant=True`. If `constant=False`, just do a
            single coefficient per CV.
        """
        self.create_maps(map_neval=map_neval)
        self.get_is_cv_values(jac_neval=jac_neval)
        if self.cv_values:
            # only run if we are using control variates
            self.get_weight_prime(constant=constant)
        self.garbage_collect()

    def garbage_collect(self, memory: Optional[str] = None) -> None:
        """
        Deletes attributes depending on choices of self.memory to clear up memory.

        Parameters:
        memory (default None) - A measure of how many attributes to delete to clear up
            memory. Can be 'max', 'large', 'medium' or 'tiny'. More info in the class
            __init__ docstring.
        """
        memory = memory or self.memory

        # Initialize properties so result is saved before deleting arrays they need
        self.stdev
        self.mean
        self.w_stdev
        self.w_mean

        # Large arrays not used in anything
        if memory in {"large", "medium", "tiny"}:
            self._delete("xs", "is_jac")
        # Large arrays but used in properties below
        if memory in {"medium", "tiny"}:
            self._delete("weight_prime", "weight_value")
        # All bigger than a certain threshold
        if memory == "tiny":
            attr_items = list(self.__dict__.items())
            for attr, attr_val in attr_items:
                if getsizeof(attr_val) > self.TINY_THRESHOLD:
                    self.__delattr__(attr)

    def _delete(self, *attrs):
        """Delete attribute if it still exists as one."""
        for attr in attrs:
            if attr in self.__dir__():
                self.__delattr__(attr)

    @property
    @check_attrs("weight_prime", "jac_neval")
    def stdev(self) -> float:
        """Standard deviation of CV function"""
        self._stdev = np.std(self.weight_prime) / np.sqrt(self.jac_neval)
        return self._stdev

    @property
    @check_attrs("weight_value", "jac_neval")
    def w_stdev(self) -> float:
        """Standard deviation of IS function"""
        self._w_stdev = np.std(self.weight_value) / np.sqrt(self.jac_neval)
        return self._w_stdev

    @property
    @check_attrs("stdev")
    def var(self) -> float:
        """Variance of CV function"""
        self._var = self.stdev**2
        return self._var

    @property
    def w_var(self) -> float:
        """Variance of IS function"""
        self._w_var = self.w_stdev**2
        return self._w_var

    @property
    @check_attrs("weight_prime")
    def mean(self) -> float:
        """Mean of CV function"""
        self._mean = np.mean(self.weight_prime)
        return self._mean

    @property
    @check_attrs("weight_value")
    def w_mean(self) -> float:
        """Mean of IS function"""
        self._w_mean = np.mean(self.weight_value)
        return self._w_mean

    @property
    def vpr(self) -> float:
        """
        Variance percentage reduction, i.e. by what percent was the variance
        reduced due to the CVs.
        """
        self._vpr = 1 - self.var / self.w_var
        return self._vpr

    def compare(
        self, rounding: int = 3, cutoff: Union[int, tuple[int, int]] = 3
    ) -> None:
        """
        Print out mean, variance and standard deviation for without and with the CVs.

        Parameters:
        rounding - How many digits to round the numbers to.
        cutoff - The power (base 10) at which to switch from floating point to
            scientific notation. That is if `cutoff=3`, then any value that is less than
            10^-3 or greater than 10^-3 will be printed in scientific notation. Can also
            be given as a tuple representing the lower and upper bound separately. So if
            `cutoff=(-3, 5)`, then the value would be printed in scientific notation if
            it is less than 10^-3 or greater than 10^5.
        """
        if isinstance(cutoff, int):
            neg_co, pos_co = 10 ** (-cutoff), 10**cutoff
        else:
            neg_co, pos_co = 10 ** (-cutoff[0]), 10 ** cutoff[1]

        def vtype(val):
            """Returns 'e' for scientific method or 'f' for float"""
            if np.isnan(val) or (val > neg_co and val < pos_co):
                return "f"
            return "e"

        w_mean = f"{self.w_mean:.{rounding}{vtype(self.w_mean)}}"
        w_var = f"{self.w_var:.{rounding}{vtype(self.w_var)}}"
        w_stdev = f"{self.w_stdev:.{rounding}{vtype(self.w_stdev)}}"
        mean = f"{self.mean:.{rounding}{vtype(self.mean)}}"
        var = f"{self.var:.{rounding}{vtype(self.var)}}"
        stdev = f"{self.stdev:.{rounding}{vtype(self.stdev)}}"
        vpr = f"{100 * self.vpr:.{rounding}{vtype(100 * self.vpr)}}%"

        plural = "s" if len(self.cv_nitn) > 1 else ""
        titles = f"No CV{plural}", f"With CV{plural}"
        print(f"{9*' '}|{titles[0]:^{8 + rounding}}|{titles[1]:^{8 + rounding}}")
        print(f"---------+{'-'*(8 + rounding)}+{'-'*(8 + rounding)}")
        print(f"Mean     |{w_mean:>{7 + rounding}} |{mean:>{7 + rounding}}")
        print(f"Variance |{w_var:>{7 + rounding}} |{var:>{7 + rounding}}")
        print(f"St Dev   |{w_stdev:>{7 + rounding}} |{stdev:>{7 + rounding}}")
        print(f"VPR      |{' ' * (7 + rounding)} |{vpr:>{7 + rounding}}")



"""
Custom types for type hinting 
"""
from typing import Callable

from nptyping import Float, NDArray, Shape

# Type for input to the batchintegrand vegas functions
_x = NDArray[Shape["'* batchSize, Dim dimension'"], Float]
_f = NDArray[Shape["'* batchSize'"], Float]
# The batch integrand vegas function type
_ftype = Callable[[_x], _f]



from importlib.metadata import version

from .cvintegrator import CVIntegrator, quick_integrate
from .functions import make_func

__all__ = ["make_func", "CVIntegrator", "quick_integrate"]
__version__ = version("covvvr")



"""
This file contains a dataclass for various functions, Each class has the following
properties and attributes:
    - name: Capitalized and spaced name
    - dimension/dim: The dimension of the integral
    - true_value: The actual value of the integral
    - _function/_f: The function that is actually called by Vegas
    - function/f: The function in question to integrate. Wraps _f/_function
        for direct use.
    - And any other parameters that the integrand may have.

_function/_f are vectorized to support batch integration by Vegas, so they take in
a Numpy array of lists of length `dim`. So f/function exist to be used manually by the
user. So running:

    f(1, 2, 3) == _f(np.array([[1], [2], [3]])) or
    f([1, 2], [3, 4]) = _f(np.array([[1, 2], [3, 4]]))

to cut down on clutter.

All functions are to be integrated from 0 to 1.
"""
from dataclasses import dataclass, field, fields, make_dataclass
from math import erf
from numbers import Number
from typing import Optional, Sequence, Union

import numpy as np
from vegas import batchintegrand

from ._exceptions import ParameterBoundError
from ._types import _f, _ftype, _x


# TODO: add getters and setters?
@dataclass(repr=False)
class Function:
    """
    Base dataclass for functions. When inheriting this class, the vectorized
    function _function is defined to override the dummy function defined here
    for Vegas's batchintegrand.

    Parameters:
    dimension - The dimension of the integral
    name - Capitalized and spaced name
    true_value (default 0) - The actual value of the integral

    dim, _f and f are equivalent to dimension, _function and function but are just
    shorter.
    """

    dimension: int
    name: str
    true_value: Optional[float] = None

    def __post_init__(self):
        """Shorthand property names"""
        self.dim = self.dimension
        self._f = self._function
        self.f = self.function

    def __repr__(self):
        """Doesn't output parameters that equal None"""
        parameters = [
            f"{f.name}={getattr(self, f.name)}"
            for f in fields(self)
            if getattr(self, f.name) is not None
        ]
        return f"{type(self).__name__}({', '.join(parameters)})"

    def _function(self, x: _x) -> _f:
        """
        Function to be called by Vegas's batch integration. Defined by inheriting
        classes.
        """
        pass

    def function(self, *x: Union[float, Sequence[float]]) -> _f:
        """
        A wrapper around self._f that allows for easier manual use. Just supply however
        many data points to evaluate as individual arguments and this method will
        deal with the proper array shapes needed.
        """
        if len(x) == 1:
            if isinstance(x[0], Number):
                # If just a single number is passed
                return self._function(np.array([[x]]))[0][0]
            else:
                # If just a single tuple of `dim` numbers are passed
                return self._function(np.array([x[0]]))[0]
        # If multiple args are passed
        return self._function(np.array(np.atleast_1d(*x)))


# TODO: Allow passing function being nonvectorized, use np.vectorize or too slow?
def make_func(
    cname: str,
    dimension: int,
    function: _ftype,
    name: Optional[str] = None,
    **params: float,
) -> Function:
    """
    Create your own dataclass function. The function provided must be vectorized.
    If the integral is n dimensional, then the input to `f` is (..., n) dimensional
    where ... is a variable amount representing how many points are being evaluated
    simultaneously. The callable `f` to be passed is the function to be integrated.
    Since batchintegration is being used from Vegas, this function must be vectorized.
    So the one dimensional function, `def f(x): return x**2 + 1`, would become:

                def f(self, x):
                    return x[:, 0]**2 + 1

    using Numpy array slicing (the `self` is also necessary since this is a class
    method). For an n-dimensional integrand, one can reference the ith variable with
    x[:, i]. Or if all variables are being summed together (like for the exponent of a
    n-dimensional Gaussian), one can use `np.sum(x, axis=1)`. More information on this
    and an example can be found in the file docstring of this file, functions.py.
    Also keyword arguments for the parameters that may be used in `f`. For example,
    if we pass to this method a=3, then we can write

                def f(self, x):
                    return self.a * (x[:, 0]**2 * x[:, 1] + 1)

    to parameterize it.

    Parameters:
    cname - The class name
    dimension - The dimension of the integral
    f - The function, must be vectorized. See above.
    name (default None) - The spaced, capitalized name for the function. If not given,
        will just use whatever `cname` is.
    params - Function parameters

    Returns (
        Initialized Function class
    )
    """
    name = cname if name is None else name
    # Dataclass fields
    fields = [(param, float, field(default=val)) for param, val in params.items()]

    # Create Function class then added new stuff
    func = Function(
        name=name,
        dimension=dimension,
    )
    func.__class__ = make_dataclass(
        cname,
        fields=fields,
        bases=(Function,),
        namespace={"_function": batchintegrand(function)},
        repr=False,
    )
    func._f = func._function
    return func


@dataclass(repr=False)
class NGauss(Function):
    """
    N-dimensional Gaussian

    Parameters:
    mu (default 0.5) - Mean of Gaussian, must be between 0 and 1
    sigma (default 0.2) - Standard deviation of Gaussian
    dimension - Dimension of Gaussian
    """

    mu: float = 0.5
    sigma: float = 0.2
    name: str = field(default="{}D Gaussian", init=False)

    def __post_init__(self):
        super().__post_init__()
        if self.mu > 1 or self.mu < 0:
            raise ParameterBoundError(
                f"Mean must be between the bounds [0, 1]. Currently set at {self.mu}."
            ) from None

        self.true_value = (
            (erf((1 - self.mu) / self.sigma) + erf(self.mu / self.sigma)) / 2
        ) ** self.dim
        self.name = self.name.format(self.dim)

    @batchintegrand
    def _function(self, x: _x) -> _f:
        norm_factor = 1 / (self.sigma * np.sqrt(np.pi)) ** self.dim
        exp = -np.sum((x - 0.5) ** 2, axis=1) / self.sigma**2
        return norm_factor * np.exp(exp)


@dataclass(repr=False)
class NCamel(Function):
    """
    N-dimensional Camel (double Gaussian) Function

    Parameters:
    mu1 (default 1/3) - Mean of first Gaussian, must be between 0 and 1
    mu2 (default 2/3)- Mean of second Gaussian, must be between 0 to 1
    sigma (default 0.2) - Standard deviation of both Gaussians
    dimension - Dimension of Camel function
    """

    mu1: float = 1 / 3
    mu2: float = 2 / 3
    sigma: float = 0.2
    name: str = field(default="{}D Camel", init=False)

    def __post_init__(self):
        super().__post_init__()
        if self.mu1 > 1 or self.mu1 < 0:
            raise ParameterBoundError(
                f"First mean must be between [0, 1]. Currently set at {self.mu1}."
            ) from None
        if self.mu2 > 1 or self.mu2 < 0:
            raise ParameterBoundError(
                f"Second mean must be between [0, 1]. Currently set at {self.mu2}."
            ) from None

        self.true_value = (
            (
                erf((1 - self.mu1) / self.sigma)
                + erf(self.mu1 / self.sigma)
                + erf((1 - self.mu2) / self.sigma)
                + erf(self.mu2 / self.sigma)
            )
            / 4
        ) ** self.dim
        self.name = self.name.format(self.dim)

    @batchintegrand
    def _function(self, x: _x) -> _f:
        norm_factor = 1 / (2 * (self.sigma * np.sqrt(np.pi)) ** (self.dim))
        exp1 = -np.sum((x - self.mu1) ** 2, axis=1) / self.sigma**2
        exp2 = -np.sum((x - self.mu2) ** 2, axis=1) / self.sigma**2
        return norm_factor * (np.exp(exp1) + np.exp(exp2))


@dataclass(repr=False)
class EntangledCircles(Function):
    """
    A function of two shifted circles. True value depends on default values.

    Parameters:
    p1, p2 (defaults 0.4, 0.6) - Shifts of the two variables
    r (default 0.25) - Radius
    w, a (default 1/0.004, 3)
    """

    p1: float = 0.4
    p2: float = 0.6
    r: float = 0.25
    w: float = 1 / 0.004
    a: float = 3
    dimension: int = field(default=2, init=False)
    name: str = field(default="Entangled Circles", init=False)

    def __post_init__(self):
        super().__post_init__()
        self.true_value = 0.01368

    @batchintegrand
    def _function(self, x: _x) -> _f:
        x1 = x[:, 0]
        x2 = x[:, 1]
        exp1 = np.exp(
            -self.w * abs((x2 - self.p2) ** 2 + (x1 - self.p1) ** 2 - self.r**2)
        )
        exp2 = np.exp(
            -self.w
            * abs((x2 - 1 + self.p2) ** 2 + (x1 - 1 + self.p1) ** 2 - self.r**2)
        )
        return x2**self.a * exp1 + (1 - x2) ** self.a * exp2


@dataclass(repr=False)
class AnnulusWCuts(Function):
    """
    An annulus with hard cuts

    Parameters:
    rmin - Radius of inner circle
    rmax - Radius of outer circle
    """

    rmin: float = 0.2
    rmax: float = 0.45
    dimension: int = field(default=2, init=False)
    name: str = field(default="Annulus with Cuts", init=False)

    def __post_init__(self):
        super().__post_init__()
        self.rminsq = self.rmin**2
        self.rmaxsq = self.rmax**2
        if self.rmin < 0:
            raise ParameterBoundError(
                f"Minimum radius must be positive. Currently set at {self.rmin}."
            ) from None
        if self.rmax < 0:
            raise ParameterBoundError(
                f"Maximum radius must be positive. Currently set at {self.rmax}."
            ) from None
        if self.rmin > self.rmax:
            raise ParameterBoundError(
                "Minimum radius must be less than maximum radius but they are "
                + f"{self.rmin} and {self.rmax} currently."
            ) from None
        self.true_value = np.pi / 4 * (self.rmax**2 - self.rmin**2)

    @batchintegrand
    def _function(self, x: _x) -> _f:
        dist = x[:, 0] ** 2 + x[:, 1] ** 2
        return np.logical_and(dist > self.rminsq, dist < self.rmaxsq).astype(float)


@dataclass(repr=False)
class ScalarTopLoop(Function):
    """
    A one-loop scalar box integral. True value depends on default values.

    Parameters:
    s12, s23, s1, s2, s3, s4 (defaults 130**2, -130**2, 0, 0, 0, 125**2) =
        Parameters of function
    mtsq (default 173.9**2) - Square of top quark mass
    """

    s12: float = 130**2
    s23: float = -(130**2) / 2
    s1: float = 0
    s2: float = 0
    s3: float = 0
    s4: float = 125**2
    mtsq: float = 173.9**2
    dimension: int = field(default=3, init=False)
    name: str = field(default="Scalar Top Loop", init=False)

    def __post_init__(self):
        super().__post_init__()
        self.true_value = 1.93741e-10

    def _Fbox(self, x, s12, s23, s1, s2, s3, s4):
        m1sq, m2sq, m3sq, m4sq = self.mtsq, self.mtsq, self.mtsq, self.mtsq
        return (
            -s12 * x[:, 1]
            - s23 * x[:, 0] * x[:, 2]
            - s1 * x[:, 0]
            - s2 * x[:, 0] * x[:, 1]
            - s3 * x[:, 1] * x[:, 2]
            - s4 * x[:, 2]
            + (1 + x[:, 0] + x[:, 1] + x[:, 2])
            * (x[:, 0] * m1sq + x[:, 1] * m2sq + x[:, 2] * m3sq + m4sq)
        )

    def _Sbox(self, x, s12, s23, s1, s2, s3, s4):
        return 1 / self._Fbox(x, s12, s23, s1, s2, s3, s4) ** 2

    @batchintegrand
    def _function(self, x: _x) -> _f:
        return (
            self._Sbox(x, self.s12, self.s23, self.s1, self.s2, self.s3, self.s4)
            + self._Sbox(x, self.s23, self.s12, self.s2, self.s3, self.s4, self.s1)
            + self._Sbox(x, self.s12, self.s23, self.s3, self.s4, self.s1, self.s2)
            + self._Sbox(x, self.s23, self.s12, self.s4, self.s1, self.s2, self.s3)
        )


@dataclass(repr=False)
class NPolynomial(Function):
    """
    N-dimensional Polynomial of the form: sum of -x_i^2 + x_i.

    Parameters:
    dimension - Dimension of Polynomial
    """

    name: str = field(default="{}D Polynomial", init=False)

    def __post_init__(self):
        super().__post_init__()
        self.true_value = self.dim / 6
        self.name = self.name.format(self.dim)

    @staticmethod
    @batchintegrand
    def _function(x: _x) -> _f:
        return np.sum(-(x**2) + x, axis=1)
